#include <stdio.h>
#include <pwd.h>
#include <dirent.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <sys/procfs.h>

#define  ON	1
#define  OFF	0

typedef struct process_table_t {
	int
		pid,
		tty;
	char	cpu_time[10];
	char	command[50];
} proc_t;

int	option_user = OFF;
int	option_default = OFF;
char	*user_name = "";

extern	int optind;

main(int argc, char *argv[])
{
	void	fatal(char *);
	proc_t	**read_process_table(DIR *);
	void	print_process_table(proc_t **);
	void	free_table(proc_t **);

	DIR	*proc_directory;
	proc_t	**process_table;
	int	options = 0;


	while((options = getopt(argc, argv, "u")) != EOF)
	{
	switch(options)
		{
		case	'u' : option_user = ON;
			user_name = *(argv + optind);
			break;
		}
	}

	if(options == OFF)
		option_default = ON;

	if(!(proc_directory = opendir("/proc")))
		fatal("opendir");

	process_table = read_process_table(proc_directory);
	print_process_table(process_table);
	free_table(process_table);	

	closedir(proc_directory);
}

proc_t **
read_process_table(DIR *proc_directory)
{
	proc_t	*read_proc_directory(DIR *);
	proc_t	**process_table;

	proc_t	*buff;
	int	count = 0;

	buff = (proc_t *)1;

	if(buff)
	{
	do	{
		if(count == 0)
			process_table = (proc_t **)malloc(sizeof(proc_t *));
		else	if((process_table = 
			(proc_t **)realloc(process_table, (count + 1)*sizeof(proc_t *))) == NULL)
				{
				puts("realloc");
				exit(0);
				}
		buff = read_proc_directory(proc_directory);

		if(buff)
			process_table[count++] = buff;
		} while(buff);

	if(count == 0) count = 1;
	process_table[count - 1] = (proc_t *)malloc(sizeof(proc_t *));
	process_table[count - 1]->pid = -1;

	}
		
	return process_table;
	
}

proc_t *
read_proc_directory(DIR *proc_directory)
{
	void	fatal(char *);
	void	calculate_time(time_t, char *);

	struct	passwd *password;
	struct	dirent *content;
	prpsinfo_t	prpsinfo;
	char	path[256];
	struct	stat file_state;
	int	allocated_flag = 0;
	int	fd;
	proc_t	*process;
	dev_t	dev_tty;
	dev_t	current_tty;


next :
	while((content = readdir(proc_directory)) &&
		(*(content->d_name) < '0' || *(content->d_name) > '9'));
	if(!content || !content->d_name)
		return NULL;

	sprintf(path, "/proc/%s", content->d_name);
	if(stat(path, &file_state) < 0)
		goto next;

	if(!allocated_flag)
		{
		process = (proc_t *)calloc(1, sizeof(*process));
		allocated_flag = 1;
		}
	if((fd = open(path, O_RDONLY)) < 0)
		{
		close(fd);
		goto next;
		}
	if(ioctl(fd, PIOCPSINFO, &prpsinfo) < 0)
		fatal("ioctl");

	dev_tty = (prpsinfo.pr_ottydev) & 0377;
	current_tty = atoi((strrchr(ttyname(0), '/') + 1));

	if(option_user != ON)
		{
		if(dev_tty != current_tty) 
			goto next;
		}
	else	if((password = getpwnam(user_name)) == NULL)
			fatal("getpwnam");
	else	if(password->pw_uid != prpsinfo.pr_uid) 
			return NULL;

	process->pid = prpsinfo.pr_pid;
	process->tty = (prpsinfo.pr_ottydev) & 0377;
	if(process->tty == 255)
		process->tty = 1;
	calculate_time(prpsinfo.pr_time.tv_sec, process->cpu_time);
	strcpy(process->command, prpsinfo.pr_psargs);

	close(fd);
	return process;
}
void
calculate_time(time_t second, char *time_buff)
{
	int	days, hours, minutes, seconds;
	char	day[4], hour[3], minute[3];


	seconds = second;

	days = seconds / 86400;
	seconds = seconds % 86400;
	hours = seconds / 3600;
	seconds = seconds % 3600;
	minutes = seconds / 60;
	seconds = seconds % 60;
	
	if(days)
		sprintf(time_buff, "%d ", days);
	else	strcpy(time_buff, "");

	if(hours)
		sprintf(hour, "%d:", hours);
	else	strcpy(hour, "0:");

	strcat(time_buff, hour);
	if(minutes)
		sprintf(minute, "%02d", minutes);
	else	strcpy(minute, "00");
	strcat(time_buff, minute);

}

void
print_process_table(proc_t **process)
{
	int	i;


	if(process[0]->pid != -1)
		printf("%6s %-8s %-4s %-20s\n", "PID", "TTY", "TIME", "CMD");
	for(i = 0;process[i]->pid > 0;process++)
		{
		printf("%6d ", process[i]->pid);
		printf("pts/%-3d ", process[i]->tty);
		printf("%5s ", process[i]->cpu_time);
		printf("%-20s\n", process[i]->command);
		}
}

void
free_table(proc_t **process_table)
{
	void	free_process(proc_t *);


	proc_t	**process;
	int	i = 0;


	for(process = process_table;*process;process++, i++)
		if((*process)->pid == -1)
			{
			free_process(*process);
			free(process_table);
			return;
			}
	
	free(process_table);
}

void
free_process(proc_t *process)
{
	if(!process) return;
	free(process);
}

void
fatal(char *error_name)
{
	perror(error_name);
	exit(1);
}
